class NumpyizeConstants:
  COLUMN_SERIES_INDEX = 'series_index'
  COLUMN_TIME_ID = 'time_id'

  SPARSE_VALUES = 'values'
  SPARSE_ROW_IND = 'row_ind'
  SPARSE_COL_IND = 'col_ind'

  SPARSE_MAPPING = 'mapping'
  SPARSE_FREQ = 'freq'
  SPARSE_DEFAULTS = [SPARSE_MAPPING, SPARSE_FREQ]

  KEY_DATASET_ORIGIN = 'dataset_origin'
  KEY_DATASET = 'dataset'
  KEY_DATASET_TYPE = 'type'
  KEY_DATASET_FN_DATA = 'fn_data'
  KEY_DATASET_ABSOLUTE_PATH_FN_DATA = 'absolute_path_fn_data'
  KEY_DATASET_FILENAME = 'filename'

  DATASET_TRAIN = 'train'
  DATASET_TEST = 'test'
  DATASET_OTHER = 'other'
  VALID_DATASET_NAMES = [DATASET_TRAIN, DATASET_TEST, DATASET_OTHER]

  KEY_AGGREGATION_FIELDS = 'AGGREGATION_FIELDS'
  KEY_TIME_FIELD = 'TIME_FIELD'
  KEY_TARGET_FIELD = 'TARGET_FIELD'
  KEY_OTHER_COVARIATE_FIELDS = 'OTHER_COVARIATE_FIELDS'
  KEY_FREQ = 'FREQ'
  KEY_NR_SERIES = 'NR_SERIES'
  KEY_NR_STEPS = 'NR_STEPS'
  KEY_TARGET_STEPS = 'TARGET_STEPS'
